/*
 * abissd/sched_basic.c - abissd, "basic" scheduler
 *
 * Written 2004, 2005 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004, 2005. All rights
 * reserved. For licensing and warranty information, see the file
 * ../../COPYING.GPLv2
 */


#include <stdio.h>
#include <string.h>
#include <errno.h>

#include <abiss_msg_basic.h>
#include <abissd_msg_basic.h>
#include <abissd.h>


/*
 * The worst-case time between initiating an I/O operation and the time when
 * the first page becomes available, not taking into account I/O errors.
 *
 * We get roughly 20-30 ms for the maximum disk access time for modern disks,
 * so we use 30 ms.
 *
 * Then we need to consider everything that may be sitting in front of us in
 * the elevator. That's one maximum disk access time plus the transfer time for
 * each batch. Each stream should only have at most one batch in the queue at a
 * any given time. Now, we could do the proper math, but for simplicity, we use
 * just a batch size of 40 kB, a disk transfer rate of 20 MB/sec, and five
 * concurrent flows. So that's 5*(30 ms+2 ms) = 160 ms
 *
 * Next, we add the maximum delay the elevator may inflict upon us. Assuming
 * that the whole high-priority work fits into a single read phase, that would
 * be the duration of the write phase. Thus we get an additional 33 ms.
 *
 * @@@ FIXME: writes need further adjusting !
 *
 * Last but not least, there may be a low-priority request at the head of the
 * queue. We need to take this one into account, if it could take longer than
 * the write phase. This is fortunately not the case, given above assumptions.
 *
 * So the final latency is 30 ms+160 ms+33ms, which we round up to 250 ms.
 *
 * We ignore the actual transfer time for the entire page, since it is small
 * in comparison to the other factors.
 */

#define DEFAULT_IO_LATENCY_SECS		0.250

/*
 * The maximum latency between the time when we could theoretically initiate
 * the next I/O operation, and the time when we're actually capable of doing
 * so in real life.
 *
 * 500 ms for the time between expiration of the timer, and the time we get to
 * do the I/O in real life. This includes scheduling of a work queue job.
 * @@@ NOTE: this value was found during experiments quite a while ago, and may
 * be overly pessimistic.
 *
 * 10 ms for the difference of an ideal timer and real-life timers, ticking at
 * a rate of HZ, which can typically be as low as 100 Hz. (Note: this is
 * becoming variable with 2.6.13.)
 */

#define DEFAULT_SUBMIT_LATENCY_SECS	0.510


/*
 * The minimum number of bytes we want to read in one go. This should be at
 * least one page.
 */

#define DEFAULT_BATCH_BYTES		40960


extern struct sched_ops sched_ops;


static void init_vars(void)
{
    var_register(&sched_ops,"batch_bytes",'i',"read batch size");
    var_register(&sched_ops,"allow_batch_bytes_override",'b',
      "application can set batch bytes");
    var_register(&sched_ops,"io_latency_secs",'f',"I/O execution latency");
    var_register(&sched_ops,"submit_latency_secs",'f',"I/O submission latency");
    var_register(&sched_ops,"resize_bytes",'v',"Resize values, for testing");

    var_set_float("io_latency_secs",DEFAULT_IO_LATENCY_SECS);
    var_set_float("submit_latency_secs",DEFAULT_SUBMIT_LATENCY_SECS);
}


static void print_params(struct abissd_open_msg *msg,int from_user)
{
    struct abiss_sched_basic_prm prm;

    if (!from_user)
	fprintf(stderr,"  (expecting %d)\n",
	  (int) sizeof(struct abiss_sched_basic_prm));

    memcpy(&prm,msg+1,sizeof(prm));

    fprintf(stderr,"  ra_bytes         %d\n",(int) prm.ra_bytes);
    fprintf(stderr,"  fill_Bps         %d\n",prm.fill_Bps);
    fprintf(stderr,"  batch_bytes      %d%s\n",(int) prm.batch_bytes,
      from_user && !var_is_set("allow_batch_bytes_override") ?
      " (ignored)" : "");
    if (!from_user)
	fprintf(stderr,"  max_credit_bytes %d\n",prm.max_credit_bytes);
}


static int open_stream(struct stream *stream,struct abissd_open_msg *msg)
{
    struct abiss_sched_basic_prm *_prm =
      (struct abiss_sched_basic_prm *) (msg+1);
    struct abiss_sched_basic_prm prm;
    double io_latency = var_get_float("io_latency_secs");
    double submit_latency = var_get_float("submit_latency_secs");

    memcpy(&prm,_prm,sizeof(prm));

    /*
     * We grow the buffer to allow for the I/O to trickle in asynchronously
     * (io_latency).
     *
     * Then we need to allow for any delays before I/O actually begins. This
     * is submit_latency plus any meta-data I/O. For simplicity, we
     * pessimisitically  assume meta-data I/O to be independent of the real
     * data I/O, so we get another io_latency.
     *
     * Last but not least, we add one, to protect against rounding in the
     * floating-point multiplication.
     */
    prm.ra_bytes += (2*io_latency+submit_latency)*prm.fill_Bps+1;

    stream->rm.byterate = prm.fill_Bps;
    stream->rm.bufsize = prm.ra_bytes;
    if (!prm.batch_bytes || !var_is_set("allow_batch_bytes_override"))
	prm.batch_bytes = var_is_set("batch_bytes") ?
	  var_get_int("batch_bytes") : DEFAULT_BATCH_BYTES;

    /*
     * Ideally, we get scheduled when we're allowed to read the first byte
     * after the batch. So the maximum delay caused by batching is
     * batch_bytes-1.
     *
     * Furthermore, we have the system's scheduling latency to consider.
     *
     * To this, we add any time we spend waiting for meta-data I/O. This can be
     * as bad as io_latency again.
     *
     * Last but not least, we add one, to protect against rounding in the
     * floating-point multiplication.
     */
    prm.max_credit_bytes =
      prm.batch_bytes-1+(submit_latency+io_latency)*prm.fill_Bps+1;

    memcpy(_prm,&prm,sizeof(prm));

    if (verbose) {
	fprintf(stderr,"resulting parameters:\n");
	print_params(msg,0);
    }

    return rm_reserve(&stream->rm) ? 0 : -EDQUOT;
}


static void close_stream(struct stream *stream)
{
    rm_return(&stream->rm);
}


static void notify(struct stream *stream,struct abissd_notification_msg *msg)
{
    struct abissd_sched_basic_n_msg *m =
      (struct abissd_sched_basic_n_msg *) msg;

    if (m->type != abissd_sb_progress) {
	fprintf(stderr,"ignoring unknown notification type %d\n",m->type);
	return;
    }
    if (verbose)
	fprintf(stderr,"notification: ID %lu status %d\n",
	    (unsigned long) m->u.progress.id,m->u.progress.status);
}


static void handle_signal(struct stream *stream)
{
    static int nth = 0;
    struct abissd_sched_basic_u_msg msg = {
	.type = abissd_sb_resize,
    };
    int size;

    if (!var_is_set("resize_bytes") || !var_get_vec("resize_bytes",nth,&size))
	return;
    msg.u.resize.id = nth;
    msg.u.resize.size_bytes = size;
    if (verbose)
	fprintf(stderr,"requesting resize: ID %lu size %d\n",
	  (unsigned long) nth,size);
    send_unsolicited(stream,
      (struct abissd_unsolicited_msg *) &msg,sizeof(msg));
    nth++;
}


struct sched_ops sched_ops = {
    .name = "basic",
    .init = init_vars,
    .print_params = print_params,
    .open = open_stream,
    .close = close_stream,
    .notify = notify,
    .signal = handle_signal,
};

PLUGIN_LICENSE("GPLv2");
