/*
 * abissd/abissd.c - ABISS user-space daemon
 *
 * Written 2004-2005 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004-2005. All rights
 * reserved. For licensing and warranty information, see the file
 * ../../COPYING.LGPLv21
 */

#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <assert.h>
#include <sys/types.h>

#include <abiss_msg.h>
#include <abissd_msg.h>

#include "common.h"
#include "rm.h"
#include "var.h"
#include "stream.h"


int verbose = 0;

static int dev_fd;


static void daemonize(void)
{
    pid_t pid;

    pid = fork();
    if (pid < 0) {
	perror("fork");
	exit(1);
    }
    if (pid) {
	if (verbose)
	    fprintf(stderr,"\nABISS daemon running with pid %d\n\n",(int) pid);
	exit(0);
    }
    if (setsid() < 0) {
	perror("setsid");
	exit(1);
    }
    (void) close(0);
    (void) close(1);
    /* keep stderr available for progress and error reporting */
    (void) chdir("/");
}


static void usage(const char *name)
{
    fprintf(stderr,
"usage: %s "
          "[-d device] [-v [-v]] [-Y ...|-N ...]\n"
"          [-n numfiles] [-a byterate] [-s bufsize] [-o var=value ...]\n\n"
"  -d device    use the specified device, instead of /dev/misc/abissd\n"
"  -v           be slightly verbose\n"
"  -v -v        be very verbose\n"
"  -Y           accept request(s), don't require any further configuration\n"
"  -N           deny request(s) with EDQUOTA\n"
"  -n numfiles  maximum number of open files\n"
"  -a byterate  maximum accumulated byterate\n"
"  -s bufsize   maximum total buffersize\n\n"
"  -o var=value set a configuration variable (see below)\n\n"
"The following variables are recognized:\n",name);
    var_usage(stderr);
    exit(1);
}


void send_unsolicited(const struct stream *stream,
  struct abissd_unsolicited_msg *msg,size_t size)
{
    struct abissd_unsolicited_msg dummy_msg,*_msg = &dummy_msg;
    ssize_t wrote;

    if (msg)
	_msg = msg;
    else
	size = sizeof(*msg);
    _msg->type = abissd_unsolicited;
    _msg->id = stream->id;
    wrote = write(dev_fd,_msg,size);
    if (wrote < 0) {
	if (errno == EIDRM) {
	    if (verbose)
		fprintf(stderr,"identifier is being removed\n");
	    return;
	}
	perror("write");
	exit(1);
    }
    if (wrote != size) {
	fprintf(stderr,"short write (%d < %d)\n",(int) wrote,(int) size);
	exit(1);
    }
    if (verbose)
	fprintf(stderr,"sent an unsolicited message\n");
}


static void handle_sigusr1(int sig)
{
    struct stream *stream;

    for (stream = streams; stream; stream = stream->next)
	if (stream->sched_ops->signal)
	    stream->sched_ops->signal(stream);
}


static void chat(void)
{
    struct sigaction sigact;
    sigset_t sigset;
    static void *buffer = NULL;
    static int buffer_size = 1024;
    ssize_t got,wrote;
    enum abissd_msg type;
    size_t ret;
    int error;

    sigact.sa_handler = handle_sigusr1;
    if (sigemptyset(&sigact.sa_mask) < 0) {
	perror("sigemptyset");
	exit(1);
    }
    sigact.sa_flags = 0;
    if (sigaction(SIGUSR1,&sigact,NULL) < 0) {
	perror("sigaction");
	exit(1);
    }
    if (sigemptyset(&sigset) < 0) {
	perror("sigemptyset");
	exit(1);
    }
    if (sigaddset(&sigset,SIGUSR1) < 0) {
	perror("sigaddset");
	exit(1);
    }
    if (!buffer) {
again:
	buffer = realloc(buffer,buffer_size);
	if (!buffer) {
	    perror("realloc");
	    exit(1);
	}
    }
again_signal:
    if (sigprocmask(SIG_UNBLOCK,&sigset,NULL) < 0) {
	perror("sigprocmask(SIG_UNBLOCK)");
	exit(1);
    }
    got = read(dev_fd,buffer,buffer_size);
    if (sigprocmask(SIG_BLOCK,&sigset,NULL) < 0) {
	perror("sigprocmask(SIG_BLOCK)");
	exit(1);
    }
    error = errno;
    if (got < 0) {
	if (error == EINTR)
	    goto again_signal;
	if (error == EMSGSIZE) {
	    buffer_size <<= 1;
	    goto again;
	}
	fprintf(stderr,"read: %s\n",strerror(error));
	exit(1);
    }
    if (!got) {
	fprintf(stderr,"EOF from daemon\n");
	exit(1);
    }
    type = *(enum abissd_msg *) buffer;
    if (verbose) {
	fprintf(stderr,"received %d bytes\n",(int) got);
	if (verbose > 1)
	    fprintf(stderr,"message type %d (%s)\n",type,
	      type == abissd_open_req ? "open_req" :
              type == abissd_close_ind ? "close_ind" :
              type == abissd_disk_suspend_req ? "suspend_req" :
              type == abissd_disk_resume_req ? "resume_req" :
	      type == abissd_notification ? "notification" :
              "unknown");
    }
    switch (type) {
	case abissd_open_req:
	    ret = abissd_open(buffer);
	    break;
	case abissd_close_ind:
	    ret = abissd_close(buffer);
	    break;
	case abissd_notification:
	    ret = abissd_notify(buffer);
	    break;
	default:
	    fprintf(stderr,"unrecognized message type %d\n",type);
	    exit(1);
    }
    if (!ret)
	return;
    assert(ret <= got);
    wrote = write(dev_fd,buffer,ret);
    if (wrote < 0) {
	perror("write");
	exit(1);
    }
    if (wrote != ret) {
	fprintf(stderr,"short write (%d < %d)\n",(int) wrote,(int) ret);
	exit(1);
    }
    if (verbose)
	fprintf(stderr,"sent a reply\n");
}


static void init_vars(void)
{
    static int initialized = 0;
    const char *path;

    if (initialized)
	return;

    path = getenv("ABISS_LIB_PATH");
    load_schedulers(path ? path : ".");
    initialized = 1;
}


int main(int argc,char **argv)
{
    const char *device = "/dev/misc/abissd";
    int c;

    while ((c = getopt(argc,argv,"b:d:vNYn:a:s:o:")) != EOF)
	switch (c) {
	    case 'd':
		device = optarg;
		break;
	    case 'v':
		verbose++;
		break;
	    case 'N':
		fixed_reply = 1;
		*reply++ = -EDQUOT;
		break;
	    case 'Y':
		fixed_reply = 1;
		*reply++ = 0;
		break;
	    case 'n':
		max_allowed_num_files = atoi(optarg);
		break;
	    case 'a':
		max_accu_byterate = atoi(optarg);
		break;
	    case 's':
		max_total_bufsize = atoi(optarg);
		break;
	    case 'o':
		init_vars();
		var_assignment(optarg);
		break;
	    default:
		/* for var_usage output */
		init_vars();
		usage(*argv);
	}
    if (optind != argc)
	usage(*argv);

    init_vars();
    rm_init();

    if (verbose > 1) {
	fprintf(stderr,"ABISS Daemon"
	  "------------------------------------------------------\n");
	fprintf(stderr,"using device <%s> to access the scheduler\n",device);
	fprintf(stderr,"known schedulers:\n");
	show_sched(stderr);
	fprintf(stderr,"settings:\n");
	if (fixed_reply) {
	    int *p;

	    fprintf(stderr,"  fixed repl%s: [",
	      reply-replies == 1 ? "y" : "ies");
	    for (p = replies; p != reply; p++)
		fprintf(stderr,"%s%d",p == replies ? "" : ",",*p);
	    fprintf(stderr,"] (0=allow all)\n");
	}
	else {
	    var_show(stderr);
	}
    }

     
    if ( verbose > 1 )
	fprintf(stderr,"ABISS Daemon"
	  "------------------------------------------------------\n\n");

    dev_fd = open(device,O_RDWR);
    if (dev_fd < 0) {
	perror(device);
	return 1;
    }
    daemonize();

    while (1)
	chat();
    return 0;
}
