/*
 * src/chabiss/chabiss.c - Change ABISS scheduler and allocator
 *
 * Written 2004, 2005 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004, 2005. All rights
 * reserved. For licensing and warranty information, see the file
 * ../../COPYING.GPLv2
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/ioctl.h>

#include <abiss_msg.h>


static void usage(const char *name)
{
    fprintf(stderr,"usage: %s mount_point [scheduler_name]\n",
      name);
    fprintf(stderr,"usage: %s -q file_or_directory\n",name);
    fprintf(stderr,"\n  -q  query current settings\n");
    exit(1);
}


static void set(const char *name,int argc,const char **argv,int fd)
{
    static struct abiss_set_msg msg = { .header = { .type = abiss_set }};

    switch (argc) {
	case 3:
	    if (strlen(argv[2]) >= ABISS_NAME_LENGTH) {
		fprintf(stderr,
		  "%s: scheduler name is too long (max %d characters)\n",
		  name,ABISS_NAME_LENGTH-1);
		exit(1);
	    }
	    strcpy(msg.sched_name,argv[2]);
	    /* fall through */
	case 2:
	    break;
	default:
	    usage(name);
    }
    if (ioctl(fd,ABISS_IOCTL,&msg) < 0) {
	if (errno == ENOTBLK)
	    fprintf(stderr,"The ABISS elevator is not enabled on this "
	      "device\n");
	else if (errno == ENOENT)
	    fprintf(stderr,"There is no ABISS scheduler called \"%s\"\n",
	      argv[2]);
	else
	    perror("ABISS_IOCTL (abiss_set)");
	exit(1);
    }
}


static void query(const char *name,int argc,const char **argv,int fd)
{
    static struct abiss_set_msg msg = { .header = { .type = abiss_query }};

    if (ioctl(fd,ABISS_IOCTL,&msg) < 0) {
	perror("ABISS_IOCTL (abiss_query)");
	exit(1);
    }
    printf("scheduler: %s\n",*msg.sched_name ? msg.sched_name : "(none)");
}


int main(int argc,const char **argv)
{
    const char *name = argv[0];
    int q = 0;
    int fd;

    if (argc < 2)
	usage(*argv);
    if (*argv[1] == '-') {
	if (strcmp(argv[1],"-q"))
	    usage(*argv);
	q = 1;
	argv++;
	argc--;
    }
    if (q) {
	fd = open(argv[1],O_RDONLY);
	query(name,argc,argv,fd);
    }
    else {
	struct stat st_dot,st_dotdot;

	if (chdir(argv[1]) < 0) {
	    perror(argv[1]);
	    return 1;
	}
	fd = open(".",O_RDONLY);
	if (fstat(fd,&st_dot) < 0) {
	    perror("fstat");
	    return 1;
	}
	if (stat("..",&st_dotdot) < 0) {
	    perror("stat(\"..\")");
	    return 1;
	}
	if (st_dot.st_dev == st_dotdot.st_dev &&
	  st_dot.st_ino != st_dotdot.st_ino) {
	    fprintf(stderr,"%s: not a mount point\n",argv[1]);
	    return 1;
	}
	set(name,argc,argv,fd);
    }
    (void) close(fd);
    return 0;
}
