#!/bin/bash
#
# src/tools/linuxenv.bash - Set up ABISS environment under UML, then run script
#
# Copyright (c) 2004-2006 Koninklijke Philips Electronics N.V.
#
# This program is free software; you can redistribute it and/or modify it
# under the terms of version 2 of the GNU General Public License as published
# by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# version 2 for more details.
#
# Written 2004-2006 by Werner Almesberger
#

#
# Usage:
#
ABISSD_ARGS=
# . linuxenv.bash fat_fs mount_point [scheduler]
#
# fat_fs	is a file or device containing a FAT file system
# mount_point	if the directory on which "fat_fs" will be mounted (under UML)
# scheduler	is the name of the ABISS scheduler to enable. If omitted, we
#		use "basic".
#
# Note that this script must be sourced and must be called linuxenv.bash !!!
# Also, this only works if the caller is also a bash script.
#
# The script calling linuxenv.bash must have the following structure:
#
# #!/bin/sh
# <setup before entering UML>
# ... linuxenv.bash ...
# <stuff to do in UML>
#
# WARNING: DO NOT USE THE WORD linuxenv.bash ANYWHERE ELSE IN THE FILE !
#
# After this script finishes, the environment will have the following state:
#
# - we're now running under UML
# - the host's file system is mounted read-write on /
# - "fat_fs" is mounted read-write on "mount_point", as a "msdos" file system
# - a tmpfs is mounted on /tmp
# - our current directory has not changed
# - all variables and functions defined before sourcing linuxenv.bash are
#   available
# - for convenience, $ABISS_PREFIX/bin has been added to the PATH
# - ABISS is ready for use
#

#
# This part runs on the host
#

if [ -z "$HOST_ENV" ]; then
    __self=
    case "$0" in
	/*)	__self=$0;;
	*/*)	__self=`pwd`/$0;;
	*)	# crappy [t]csh and zsh don't fix $0 for us, so we have to
		# repeat the lookup manually.
		for n in `echo $PATH | tr : ' '`; do
		    if [ -x "$n/$0" ]; then
			if [ -z "${n#/*}" ]; then
			    __self=$n/$0
			else
			    __self=`pwd`/$n/$0
			fi
			break
		    fi
		done;;
    esac
    if [ -z "$__self" -o ! -x "$__self" ]; then
	echo "can't find myself ($0)" 1>&2
	exit 1
    fi

    if [ -z "$2" ]; then
	echo "usage: linuxenv.bash fat_fs mount_point [scheduler [allocator]]" \
	  1>&2
	rm -f $__env_file
	exit 1
    fi
    if [ ! -w $1 ]; then
	echo "file system file $1 is not writeable" 1>&2
	rm -f $__env_file
	exit 1
    fi
    if [ ! -d $2 ]; then
	echo "mount point $2 is not a directory" 1>&2
	rm -f $__env_file
	exit 1
    fi

    __env_file=${TMPDIR:-/tmp}/.linuxenv.$$
    cat >$__env_file <<EOF
ABISS_PREFIX=$ABISS_PREFIX
ABISS_TOOLS=$ABISS_TOOLS
host_cwd=`pwd`
host_path=$PATH
mount_point=$2
chabiss_args="${3:-basic} $4"
abissd_args="$ABISSD_ARGS"
export ABISS_PREFIX ABISS_TOOLS
EOF
    #
    # DCL deja-vu ...
    #
    declare |
      sed '/^\('"`readonly | sed '/^[^=]* \([^=]*=\).*/s//\1\\\\|/p;d' |
        tr -d '\012'`"'=\)/d' >>$__env_file
    export >>$__env_file

    exec $RLINUX_UML_KERNEL rw \
      root=/dev/root rootflags=/ rootfstype=hostfs \
      append="log_buf_len=10M" \
      ubd0=$1 \
      HOST_ENV=$__env_file \
      init=$ABISS_TOOLS/linuxenv.bash
fi

#
# This part runs in UML
#

. $HOST_ENV
rm -f $HOST_ENV
PATH=$host_path:$ABISS_PREFIX/bin
mount -t proc proc /proc
mount -t sysfs sysfs /sys
mount -t tmpfs tmpfs /tmp

# since we don't have devfs anymore and don't want to mess up our host's /dev,
# we create our special devices in /tmp (tmpfs)
mknod /tmp/ubd0 b `tr : ' ' </sys/block/ubda/dev `
mknod /tmp/abissd c 10 `sed '/ abissd$/s///p;d' </proc/misc`

abissd -d /tmp/abissd $abissd_args
mount -t ${ABISS_FS_TYPE:-msdos} /tmp/ubd0 $mount_point
echo abiss >/sys/block/ubda/queue/scheduler
chabiss $mount_point $chabiss_args
cd $host_cwd
sed '/^#!/p;1,/\<linuxenv\.bash\>/d' <$__self >/tmp/host_script
chmod 755 /tmp/host_script
. /tmp/host_script
