/*
 * racc_linux.c - Random reader/writer, Linux AIO interface
 *
 * Written 2004 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004. All rights reserved.
 * For licensing and warranty information, see the file ../../COPYING.GPLv2
 */

/*
 * For more AIO information, see
 * http://www.kvack.org/~blah/aio/
 */


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <libaio.h>
#include <sys/types.h>
#include <sys/mman.h>

#include "racc.h"


static io_context_t ctx;


static void do_io(const char *label,struct iocb **ios,struct io_event *events,
  int nr)
{
    int i,res;

    if (verbose)
	show_time("%s setup",label);

    /* @@@ dunno if io_submit uses errno or no - assuming the worst ... */
    res = io_submit(ctx,nr,ios);
    if (res < 0) {
	fprintf(stderr,"io_submit: %s\n",strerror(-res));
	exit(1);
    }
    if (res != nr) {
	fprintf(stderr,"io_submit: submitted only %d of %d requests\n",res,nr);
	exit(1);
    }
    if (verbose)
	show_time("%s submission",label);

    if (io_getevents(ctx,nr,nr,events,NULL) < 0) {
	perror("io_getevents");
	exit(1);
    }
    if (verbose)
	show_time("%s completion",label);

    /*
     * "res" is unsigned, yet it's used ssize_t-style. Very cute :-(
     */
    for (i = 0; i != nr; i++)
	if ((events[i].res >> (sizeof(events[i].res)*8-1)) & 1)  {
	    fprintf(stderr,"request %d: %s\n",i,strerror(-events[i].res));
	    exit(1);
	}
}


void racc_linux(int fd,int flags,off_t obj_size,int access_size,
  int step_size,int accesses,int locations)
{
    struct iocb *iocb,**ios;
    struct io_event *events;
    void *buf;
    int res,i;

    /* yet another gotcha: io_setup doesn't set errno */
    res = io_setup(accesses,&ctx);
    if (res < 0) {
	fprintf(stderr,"io_setup: %s\n",strerror(-res));
	exit(1);
    }
    buf = calloc((flags == O_RDWR ? accesses : 1)*access_size+PAGE-1,1);
    if (!buf) {
	perror("calloc");
	exit(1);
    }
    buf = (void *) ((unsigned long) buf & ~(PAGE-1));
    iocb = calloc(accesses,sizeof(struct iocb));
    if (!iocb) {
	perror("calloc");
	exit(1);
    }
    ios = malloc(sizeof(struct iocb *)*accesses);
    if (!ios) {
	perror("malloc");
	exit(1);
    }
    events = malloc(sizeof(struct io_event)*accesses);
    if (!events) {
	perror("malloc");
	exit(1);
    }
    for (i = 0; i != accesses; i++) {
	ios[i] = iocb+i;
	io_prep_pread(iocb+i,fd,flags == O_RDWR ? buf+access_size*i : buf,
	  access_size,(off_t) step_size*(rand() % locations));
    }
    if (mlockall(MCL_CURRENT) < 0) {
	perror("mlockall");
	exit(1);   
    }           
    if (flags == O_RDONLY || flags == O_RDWR) {
	do_io("read",ios,events,accesses);
    }
    if (flags == O_WRONLY || flags == O_RDWR) {
	for (i = 0; i != accesses; i++)
	    iocb[i].aio_lio_opcode = IO_CMD_PWRITE;
	do_io("write",ios,events,accesses);
    }
}
