/*
 * rwrt.c - Real-time reader/writer
 *
 * Written 2004,2005 by Werner Almesberger, Giel de Nijs
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004, 2005. All rights
 * reserved. For licensing and warranty information, see the file
 * ../../COPYING.GPLv2
 */


#include <stddef.h>
#include <stdint.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <time.h>
#include <signal.h>
#include <assert.h>
#include <sched.h>
#include <sys/time.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <sys/ioctl.h>

#include <abiss_msg.h>
#include <abiss_msg_basic.h>

#include "seq.h"


#define TIMESTAMPS_MS		8
#define TIMESTAMPS_US		4
#define BARS			78
#define MAX_MLOCKED_READS	100000 /* don't try to mlock lots of memory */

struct stats {
    int rdev_us;		/* read/write deviation (in us) */
    int tdev_us;		/* read/write and playout deviation (in us) */
    int lat_us;			/* read/write latency (us) */
    int pos_us;			/* positioning latency (us) */
    int mm_us;			/* memory reclaim latency (us) */
    int slp_us;			/* sleep duration latency (us) */
    int dist_bytes;		/* positioning distance (bytes) */
    long long int run_us;	/* run time (us) */
};

enum ps_mode {
    ps_rjitter,			/* read/write jitter */
    ps_tjitter,			/* read/write jitter, with positioning */
    ps_lat,			/* read/write latency */
    ps_pos,			/* positioning latency */
    ps_mm,			/* memory reclaim latency */
    ps_slp,			/* sleep duration */
    ps_dist,			/* playout point movement distance */
};

enum print_mode {
    print_all,			/* print all stat lines */
    print_one,			/* print one stat line */
};

struct stat_options {
    int stat_time;
    int stat_hist;
    int stat_lat_time;
    int stat_lat_hist;
    int stat_pos_time;
    int stat_pos_hist;
    int stat_mm_time;
    int stat_mm_hist;
    int stat_slp_time;
    int stat_slp_hist;
    int stat_dist_time;
    int stat_dist_hist;
    int stat_overall;
    int stat_overall_hist;
};


static int unit_us = 1000; /* output unit; 1000 us */
static int timestamps = TIMESTAMPS_MS;


static void usage(const char *name)
{
    fprintf(stderr,
      "usage: %s [-v] [-s] [-o] [-x] [-M] [-m|-u]\n"
      "            [-R] [-t|-h|-c|-r|-l|-p|-a|-d|-q|-j|-k|-y|-z]\n"
      "            [-w bin_width] [-S] [-T text] file_name [[rate]\n"
      "            read/write_size [read_ahead_bytes"
      "]]\n\n"
      "  rate              read/write and fill rate (bytes per second)\n"
      "  read/write_size   read/write size (bytes)\n"
      "  read_ahead_bytes  read-ahead buffer (bytes)\n"
      "\n"
      "Sizes and rates can be followed by one of the SI prefixes \"k\" (1000)\n"
      "\"M\" (1000000), or \"G\" (1000000000). Also the respective unit\n"
      "(\"B\" or \"Bps\") can be specified, but is optional.\n"
      "\n"
      "General options:\n"
      "  -o   use SCHED_OTHER, i.e. don't raise CPU priority (and don't "
      "mlockall)\n"
      "  -x   allow excessive resource use (which may kill the system)\n"
      "  -v   verbose: print what will be, and what has been done\n"
      "  -s   silent: don't print progress indications\n"
      "  -M   enable direct reclaiming of free memory\n"
      "Statistics options:\n"
      "  -t   print jitter over time\n"
      "  -h   print histogram for jitter\n"
      "  -c   like -h, but cumulative\n"
      "  -b   print jitter of read/write and playout point update over time\n"
      "  -g   print histogram for combined jitter\n"
      "  -r   print read/write latency over time\n"
      "  -l   print histogram for read/write latency\n"
      "  -p   print playout point advancing latency over time\n"
      "  -a   print histogram for playout point advancing latency\n"
      "  -d   print direct memory reclaim duration over time\n"
      "  -q   print histogram for memory reclaim duration\n"
      "  -j   effective playout point movement distance\n"
      "  -k   print histogram for effective playout point movement distance\n"
      "  -y   print sleep duration over time (only if using a rate)\n"
      "  -z   print histogram for sleep duration\n"
      "  -m   use milliseconds (default)\n"
      "  -u   use microseconds instead of milliseconds\n"
      "  -w bin_width\n"
      "       set the histogram bin width (in microseconds)\n"
      "  -S   record or check a sequence number every 1024 bytes\n"
      "  -T text\n"
      "       embed \"text\" in the progress bar\n"
      "  -W   write instead of reading\n"
      "  -n N perform N read/write operations\n"
      "  -R   prints the output selected by -t, -b, -r, -p, -d, -j, -y\n"
      "       real-time during a test run instead of afterwards, implies -s\n"
      ,name);
    exit(1);
}

static int si(const char *name,const char *end,const char *unit)
{
    if (!*end)
	return 1;
    if (!strcmp(end,unit))
	return 1;
    if (end[1] && strcmp(end+1,unit))
	usage(name);
    switch (*end) {
	case 'k':
	    return 1000;
	case 'M':
	    return 1000*1000;
	case 'G':
	    return 1000*1000*1000;
	default:
	    usage(name);
	    return 0; /* stupid gcc */
    }
}


static void do_gettimeofday(struct timeval *tv)
{
    if (gettimeofday(tv,NULL) < 0) {
	perror("gettimeofday");
	exit(1);
    }
}


static void tv_add_us(struct timeval *a,int b)
{
    a->tv_usec += b;
    while (a->tv_usec >= 1000000) {
	a->tv_usec -= 1000000;
	a->tv_sec++;
    }
}


static void tv_sub(struct timeval *a,const struct timeval *b)
{
    a->tv_sec -= b->tv_sec;
    a->tv_usec -= b->tv_usec;
    if (a->tv_usec >= 0)
	return;
    a->tv_usec += 1000000;
    a->tv_sec--;
}


static void tv_sub_neg(struct timeval *a,const struct timeval *b)
{
    a->tv_sec = b->tv_sec-a->tv_sec;
    a->tv_usec = b->tv_usec-a->tv_usec;
    if (a->tv_usec >= 0)
	return;
    a->tv_usec += 1000000;
    a->tv_sec--;
}


static int cmp_rdev(const void *_a,const void *_b)
{
    const struct stats *a = _a,*b = _b;

    return a->rdev_us < b->rdev_us ? -1 :
      a->rdev_us > b->rdev_us ? 1 :
      a->run_us < b->run_us ? -1 :
      a->run_us > b->run_us ? 1 : 0;
}


static int cmp_tdev(const void *_a,const void *_b)
{
    const struct stats *a = _a,*b = _b;

    return a->tdev_us < b->tdev_us ? -1 :
      a->tdev_us > b->tdev_us ? 1 :
      a->run_us < b->run_us ? -1 :
      a->run_us > b->run_us ? 1 : 0;
}


static int cmp_lat(const void *_a,const void *_b)
{
    const struct stats *a = _a,*b = _b;

    return a->lat_us < b->lat_us ? -1 :
      a->lat_us > b->lat_us ? 1 :
      a->run_us < b->run_us ? -1 :
      a->run_us > b->run_us ? 1 : 0;
}


static int cmp_pos(const void *_a,const void *_b)
{
    const struct stats *a = _a,*b = _b;

    return a->pos_us < b->pos_us ? -1 :
      a->pos_us > b->pos_us ? 1 :
      a->run_us < b->run_us ? -1 :
      a->run_us > b->run_us ? 1 : 0;
}


static int cmp_mm(const void *_a,const void *_b)
{
    const struct stats *a = _a,*b = _b;

    return a->mm_us < b->mm_us ? -1 :
      a->mm_us > b->mm_us ? 1 :
      a->run_us < b->run_us ? -1 :
      a->run_us > b->run_us ? 1 : 0;
}


static int cmp_slp(const void *_a,const void *_b)
{
    const struct stats *a = _a,*b = _b;

    return a->slp_us < b->slp_us ? -1 :
      a->slp_us > b->slp_us ? 1 :
      a->run_us < b->run_us ? -1 :
      a->run_us > b->run_us ? 1 : 0;
}


static int cmp_dist(const void *_a,const void *_b)
{
    const struct stats *a = _a,*b = _b;

    return a->dist_bytes < b->dist_bytes ? -1 :
      a->dist_bytes > b->dist_bytes ? 1 :
      a->run_us < b->run_us ? -1 :
      a->run_us > b->run_us ? 1 : 0;
}


static void print_stat_hist(const struct stats *s,int n,int cumul,int sum,
  enum ps_mode mode)
{
    int i,x;

    switch (mode) {
	case ps_rjitter:
	    x = s->rdev_us/unit_us;
	    break;
	case ps_tjitter:
	    x = s->tdev_us/unit_us;
	    break;
	case ps_lat:
	    x = s->lat_us/unit_us;
	    break;
	case ps_pos:
	    x = s->pos_us/unit_us;
	    break;
	case ps_mm:
	    x = s->mm_us/unit_us;
	    break;
	case ps_slp:
	    x = s->slp_us/unit_us;
	    break;
	case ps_dist:
	    x = s->dist_bytes;
	    break;
	default:
	    abort();
    }
    printf("%5d %6d  # ",x,cumul ? sum : n);
    for (i = 0; i != timestamps && i != n; i++)
	printf("%lld%s",s[i].run_us/unit_us,i != n-1 ? "," : "");
    if (i < n-1)
	printf("...,");
    if (i != n)
	printf("%lld",s[n-1].run_us/unit_us);
    printf("\n");
}


static int field_n(const int *field_0,int n)
{
    return *((int *) ((char *) (field_0)+(n)*sizeof(struct stats)));
}


static void print_stat_times_1(const char *title,const struct stats *s,int n,
    enum print_mode pm,const int *field_0,int scale)
{
    char *__title = strdup(title);
    char *p;
    char unit_prefix = (unit_us == 1 ? 'u' : 'm');


    if (!__title) {
	perror("strdup");
	exit(1);
    }
    while ((p = strchr(__title,'$')))
	*p = unit_prefix;
    if (pm == print_one)
	    printf("%lld %d # %s\n",s[n-1].run_us/unit_us,
	      field_n(field_0,n-1)/(scale ? unit_us : 1),__title);
    else {
	int i;

	printf("# %s\n",__title);
	for (i = 0; i != n; i++)
	    printf("%lld %d\n",s[i].run_us/unit_us,
	      field_n(field_0,i)/(scale ? unit_us : 1));
	putchar('\n');
    }
}


static void print_stat_times(const struct stats *s, int n,
    const struct stat_options *st_opt, enum print_mode pm)
{

    /*
     * Prefixes used in section titles:
     *
     * Single measurement group:
     *
     * Ti   Jitter (prefix selects "Time of sample")
     * O    Overall jitter
     * Rea  Read latency (*)
     * P    Playout point advancing latency
     * D    Direct memory reclaim duration
     * S    Sleep duration
     * E    Effective playout point movement distance
     *
     * (*) Unique in first character within group. 
     */

    if (!n)
        return;

    if (st_opt->stat_time)
	print_stat_times_1("Time of sample ($s), jitter ($s)",s,n,pm,
	  &s->rdev_us,1);
    if (st_opt->stat_overall)
	print_stat_times_1("Overall jitter ($s), over time ($s)",s,n,pm,
	  &s->tdev_us,1);
    if (st_opt->stat_lat_time)
	print_stat_times_1("Read latency ($s), over time ($s)",s,n,pm,
	  &s->lat_us,1);
    if (st_opt->stat_pos_time)
	print_stat_times_1(
	  "Playout point advancing latency ($s), over time ($s)",s,n,pm,
	  &s->pos_us,1);
    if (st_opt->stat_mm_time)
	print_stat_times_1(
	  "Direct memory reclaim duration ($s), over time ($s)",s,n,pm,
	  &s->mm_us,1);
    if (st_opt->stat_slp_time)
	print_stat_times_1("Sleep duration ($s), over time ($s)",s,n,pm,
	  &s->slp_us,1);
    if (st_opt->stat_dist_time)
	print_stat_times_1(
	  "Effective playout point movement distance (bytes), over time ($s)",
	  s,n,pm,&s->dist_bytes,0);
}


static void realtimize(void)
{
    struct sched_param prm;

    if (mlockall(MCL_CURRENT | MCL_FUTURE) < 0) {
	perror("mlockall");
	exit(1);
    }
    prm.sched_priority = sched_get_priority_max(SCHED_FIFO);
    if (prm.sched_priority < 0) {
	perror("sched_get_priority_max SCHED_FIFO");
	exit(1);
    }
    if (sched_setscheduler(0,SCHED_FIFO,&prm) < 0) {
	perror("sched_setscheduler SCHED_FIFO");
	exit(1);
    }
}


static void unrealtime(void)
{
    struct sched_param prm = { .sched_priority = 0 };

    if (sched_setscheduler(0,SCHED_OTHER,&prm) < 0) {
	perror("sched_setscheduler SCHED_OTHER");
	exit(1);
    }
    if (munlockall() < 0) {
	perror("munlockall");
	exit(1);
    }
}


int main(int argc,char **argv)
{
    static struct abiss_sched_basic_prm basic_prm;
    static struct abiss_attach_msg attach_msg = {
	.header = {
	    .type = abiss_attach,
	},
    };
    static struct abiss_position_msg pos_msg = {
	.header = {
	    .type = abiss_position_fb,
	},
	.whence = SEEK_CUR,
	.pos = 0,
    };
    static struct timeval now;
    static sigset_t set;

    const char *name = *argv;
    struct stat st;
    struct timeval start,goal;
    struct stat_options st_opt;
    int stat_cumul = 0;
    int rtprint = 0;
    int unit_prefix = 'm';
    int bin_us = 1000; /* 1 ms histogram bin width */
    int verbose = 0,silent = 0,rt,raise_prio = 1,excessive = 0;
    int do_write = 0,sequence = 0,num_writes = -1;
    int read_size;
    int *fill_Bps;
    char *end;
    void *buffer;
    int fd,interval_us,n = 0,num_io,bars = 0;
    char bar_text[BARS+1];
    int seq_errors = 0;
    int direct_mm = 0;
    struct stats *stat;
    int c;

    memset(&st_opt,0,sizeof(struct stat_options));
    
    memset(bar_text,'#',BARS);
    bar_text[BARS] = 0;
    /* lower-case options still available: efi */
    while ((c = getopt(argc,argv,"abcdghjklmn:opqrstuvw:xyzMST:WR")) != EOF)
	switch (c) {
	    case 't':
		st_opt.stat_time = 1;
		break;
	    case 'h':
		st_opt.stat_hist = 1;
		break;
	    case 'c':
		stat_cumul = 1;
		break;
	    case 'b':
		st_opt.stat_overall = 1;
		break;
	    case 'g':
		st_opt.stat_overall_hist = 1;
		break;
	    case 'r':
		st_opt.stat_lat_time = 1;
		break;
	    case 'l':
		st_opt.stat_lat_hist = 1;
		break;
	    case 'o':
		raise_prio = 0;
		break;
	    case 'x':
		excessive = 1;
		break;
	    case 'p':
		st_opt.stat_pos_time = 1;
		break;
	    case 'a':
		st_opt.stat_pos_hist = 1;
		break;
	    case 's':
		silent = 1;
		break;
	    case 'v':
		verbose = 1;
		break;
	    case 'j':
		st_opt.stat_dist_time = 1;
		break;
	    case 'k':
		st_opt.stat_dist_hist = 1;
		break;
	    case 'm':
		unit_us = 1000;
		unit_prefix = 'm';
		timestamps = TIMESTAMPS_MS;
		break;
	    case 'u':
		unit_us = 1;
		unit_prefix = 'u';
		timestamps = TIMESTAMPS_US;
		break;
	    case 'w':
		bin_us = strtoul(optarg,&end,10);
		if (*end)
		    usage(name);
		break;
	    case 'n':
		num_writes = strtoul(optarg,&end,10);
		break;
	    case 'M':
		direct_mm = 1;
		break;
	    case 'd':
		st_opt.stat_mm_time = 1;
		break;
	    case 'q':
		st_opt.stat_mm_hist = 1;
		break;
	    case 'y':
		st_opt.stat_slp_time = 1;
		break;
	    case 'z':
		st_opt.stat_slp_hist = 1;
		break;
	    case 'W':
		do_write = 1;
		break;
	    case 'S':
		sequence = 1;
		break;
	    case 'T':
		{
		    int len = strlen(optarg);

		    bar_text[3] = ' ';
		    if (len > BARS-8)
			len = BARS-8;
		    strncpy(bar_text+4,optarg,len);
		    bar_text[len+4] = ' ';
		    break;
		}
	    case 'R':
		rtprint = 1;
		silent = 1;
		break;
	    default:
		usage(name);
	}

    switch (argc-optind) {
	case 4:
	    rt = 1;
	    strcpy(attach_msg.sched_name,"basic");
	    attach_msg.sched_prm = &basic_prm;
	    basic_prm.ra_bytes = strtoul(argv[optind+3],&end,0);
	    basic_prm.ra_bytes *= si(name,end,"B");
	    fill_Bps = &basic_prm.fill_Bps;
	    break;
	case 1:
	    /* fall through */
	case 2:
	    rt = 0;
	    fill_Bps = NULL;
	    break;
	case 3:
	    rt = 0;
	    fill_Bps = &basic_prm.fill_Bps; /* pick any */
	    break;
	default:
	    usage(name);
    }

    if (bin_us < unit_us)
	fprintf(stderr,"warning: bin width (%d us) < output unit (%d us)\n",
	  bin_us,unit_us);

    fd = open(argv[optind],
      do_write ? O_WRONLY | (num_writes == -1 ? 0 : O_CREAT): O_RDONLY,0666);
    if (fd < 0) {
	perror(argv[optind]);
	return 1;
    }
    if (fstat(fd,&st) < 0) {
	perror("fstat");
	return 1;
    }
    seq = st.st_ino;

    if (argc-optind == 1)
	read_size = st.st_blksize;
    else {
	read_size = strtoul(argv[optind+1+!!fill_Bps],&end,0);
	read_size *= si(name,end,"B");
	if (read_size <= 0) {
	    fprintf(stderr,"\"read/write_size\" must be greater than zero\n");
	    return 1;
	}
    }

    buffer = malloc(read_size);
    if (!buffer) {
	perror("malloc");
	return 1;
    }
    if (do_write)
	memset(buffer,'U',read_size);

    num_io = (st.st_size+read_size-1)/read_size;
    if (num_writes != -1) {
	num_io = num_writes;
	st.st_size = num_writes*read_size;
    }

    if (raise_prio && num_io > MAX_MLOCKED_READS && !excessive) {
	fprintf(stderr,"won't mlockall with %d > %d reads/writes "
          "(please use -o or -x)\n",num_io,MAX_MLOCKED_READS);
	return 0;
    }

    stat = malloc(num_io*sizeof(struct stats));
    if (!stat) {
	perror("malloc");
	return 1;
    }

    if (fill_Bps) {
	*fill_Bps = strtoul(argv[optind+1],&end,0);
	*fill_Bps *= si(name,end,"Bps");
	if (*fill_Bps <= 0) {
	    fprintf(stderr,"\"rate\" must be greater than zero\n");
	    return 1;
	}
	interval_us = read_size*1000000.0/ *fill_Bps;
                                       /* ^___________ mind the gap :-) */
	if (verbose) {
	    double t = num_io ? (num_io-1.0)*interval_us/1000000.0 : 0;

	    printf("# %s %lu bytes (%d %s operation%s) in %.6f s\n",
	      do_write ? "writing" : "reading",(unsigned long) st.st_size,
	      num_io,do_write ? "write" : "read",num_io == 1 ? "" : "s",
	      t);
	    if (t)
		printf("# (%f bytes/sec, interval %d us)\n",
		  st.st_size/t,interval_us);
	}
    }
    else {
	if (verbose)
	    printf("# greedily %s %lu bytes (%d %s operation%s)\n",
	      do_write ? "writing" : "reading",(unsigned long) st.st_size,
	      num_io,do_write ? "write" : "read",num_io == 1 ? "" : "s");
    }

    if (rt && ioctl(fd,ABISS_IOCTL,&attach_msg) < 0) {
	perror("ioctl ABISS_IOCTL (abiss_attach)");
	return 1;
    }
    if (raise_prio)
	realtimize();

    if (sigemptyset(&set) < 0) {
	perror("sigemptyset");
	return 1;
    }
    if (sigaddset(&set,SIGUSR1) < 0) {
	perror("sigaddset(SIGUSR1)");
	return 1;
    }
    if (sigprocmask(SIG_BLOCK,&set,NULL) < 0) {
	perror("sigprocmask(SIG_BLOCK,{USR1})");
	return 1;
    }

    do_gettimeofday(&start);
    goal = start;

    while (1) {
	struct timeval before_read,after_read,after_pos;
	struct timespec ts;
	int got,dt;

	do_gettimeofday(&before_read);
	if (do_write) {
	    size_t size = st.st_size > read_size ? read_size : st.st_size;

	    if (sequence)
		put_sequence(buffer,size);
	    got = write(fd,buffer,size);
	    if (got < 0) {
		perror("write");
		return 1;
	    }
	}
	else {
	    got = read(fd,buffer,read_size);
	    if (got < 0) {
		perror("read");
		return 1;
	    }
	    if (sequence) {
		if (check_sequence(buffer,got)) {
		    if (verbose)
			fprintf(stderr,"sequence error in read #%d\n",n);
		    seq_errors++;
		}
	    }
	}
	do_gettimeofday(&after_read);

	if (rt) {
	    int res;

	    res = ioctl(fd,ABISS_IOCTL,&pos_msg);
	    if (res < 0) {
		perror("ioctl ABISS_IOCTL (abiss_position_fb)");
		return 1;
	    }
	    stat[n].dist_bytes = res;
	}
	do_gettimeofday(&after_pos);

	dt = (after_read.tv_sec-goal.tv_sec)*1000000+
	  after_read.tv_usec-goal.tv_usec;
	stat[n].rdev_us = dt;

	dt = (after_pos.tv_sec-goal.tv_sec)*1000000+
	  after_pos.tv_usec-goal.tv_usec;
	stat[n].tdev_us = dt;

	if (!fill_Bps)
	    goal = after_read;

	tv_sub(&after_pos,&after_read);
	stat[n].pos_us = after_pos.tv_sec*1000000+after_pos.tv_usec;

	tv_sub(&after_read,&before_read);
	stat[n].lat_us = after_read.tv_sec*1000000+after_read.tv_usec;

	tv_sub(&before_read,&start);
	stat[n].run_us = before_read.tv_sec*1000000LLU+before_read.tv_usec;

	n++;

	if (!silent)
	    while (bars != (num_io ? n*BARS/num_io : BARS)) {
		write(2,bar_text+bars,1);
		bars++;
	    }

	do_gettimeofday(&now);

	st.st_size -= got;
	if (!st.st_size)
	    break;
	if (sigpending(&set) < 0) {
	    perror("sigpending");
	    return 1;
	}
	if (sigismember(&set,SIGUSR1))
	    break;

	if (got < read_size) {
	    fprintf(stderr,"%s(%d) returned %d, with supposedly %lu left\n",
	      do_write ? "write" : "read",read_size,got,
	      (unsigned long) st.st_size+got);
	    return 1;
	}

	if (!fill_Bps)
	    continue;

	if (direct_mm) {
		static struct abiss_msg msg = {
			.type = abiss_direct_mm,
		};
		struct timeval then = now;

		if (ioctl(fd,ABISS_IOCTL,&msg) < 0) {
			perror("ioctl ABISS_IOCTL (abiss_direct_mm)");
			exit(1);
		}
		do_gettimeofday(&now);
		tv_sub_neg(&then,&now);
		stat[n-1].mm_us = then.tv_sec*1000000LLU+then.tv_usec;
	}

	tv_add_us(&goal,interval_us);
	ts.tv_sec = goal.tv_sec-now.tv_sec;
	ts.tv_nsec = (goal.tv_usec-now.tv_usec)*1000;
	if (ts.tv_nsec < 0) {
		ts.tv_nsec += 1000*1000*1000;
		ts.tv_sec--;
	}
	stat[n-1].slp_us = ts.tv_sec*1000000+ts.tv_nsec/1000;

        if (rtprint) {
    	    print_stat_times(stat, n-1, &st_opt, print_one);
	    fflush(stdout);

	    /*
	     * print_stat_times may cause delays. We recalculate the sleep
	     * time here in an attempt to maintain the overall rate.
	     */
	    do_gettimeofday(&now);
	    ts.tv_sec = goal.tv_sec-now.tv_sec;
	    ts.tv_nsec = (goal.tv_usec-now.tv_usec)*1000;
	    if (ts.tv_nsec < 0) {
		    ts.tv_nsec += 1000*1000*1000;
		    ts.tv_sec--;
	    }
	}

	if (ts.tv_sec < 0)
	    goal = now;
	else {
	    if (nanosleep(&ts,NULL) < 0) {
		perror("nanosleep");
		return 1;
	    }
	}
    }
    if (raise_prio)
	unrealtime();
    if (!silent)
	write(2,"\n",1);

    if (verbose) {
	tv_sub(&now,&start);
	printf("# completed after %d.%06d seconds",
	  (int) now.tv_sec,(int) now.tv_usec);
	if (n != num_io)
	    printf(", did %d/%d %s",n,num_io,do_write ? "writes" : "reads");
	putchar('\n');
    }

    if (!rtprint)
    	print_stat_times(stat, n, &st_opt, print_all);

    /*
     * Prefixes used in section titles:
     * Histogram group:
     *
     * J    Jitter
     * C    Cumulative: jitter
     * To   Total jitter
     * L    Latency for read/write
     * A    Advancing (playout point) latency
     * Rec  Reclaim (free memory) latency (*)
     * H    Histogram of sleep durations
     * M    Movement distance (effective)
     *
     * (*) Unique in first character within group. 
     */

    /*
     * Do the 0...n-1 histograms first, or entry n will have been sorted into
     * the array.
     */
    if (st_opt.stat_mm_hist && n) {
	int cnt = 0,i;

	qsort(stat,n-1,sizeof(struct stats),cmp_mm);
	printf("# Reclaim (free memory) latency (%cs), "
	  "number of occurrences (per %d us)\n",unit_prefix,bin_us);
	for (i = 0; i != n-1; i++) {
	    if (i && stat[i].mm_us/bin_us != stat[i-1].mm_us/bin_us) {
		print_stat_hist(stat+i-cnt,cnt,0,0,ps_mm);
		cnt = 0;
	    }
	    cnt++;
	}
	print_stat_hist(stat+i-cnt,cnt,0,0,ps_mm);
	printf("\n");
    }
    if (st_opt.stat_slp_hist && n) {
	int cnt = 0,i;

	qsort(stat,n-1,sizeof(struct stats),cmp_slp);
	printf("# Histogram of sleep durations (%cs), "
	  "number of occurrences (per %d us)\n",unit_prefix,bin_us);
	for (i = 0; i != n-1; i++) {
	    if (i && stat[i].slp_us/bin_us != stat[i-1].slp_us/bin_us) {
		print_stat_hist(stat+i-cnt,cnt,0,0,ps_slp);
		cnt = 0;
	    }
	    cnt++;
	}
	print_stat_hist(stat+i-cnt,cnt,0,0,ps_slp);
	printf("\n");
    }
    if (st_opt.stat_hist || stat_cumul)
	qsort(stat,n,sizeof(struct stats),cmp_rdev);
    if (st_opt.stat_hist) {
	int cnt = 0,i;

	printf("# Jitter (%cs), number of occurrences (per %d us)\n",
	  unit_prefix,bin_us);
	for (i = 0; i != n; i++) {
	    if (i && stat[i].rdev_us/bin_us != stat[i-1].rdev_us/bin_us) {
		print_stat_hist(stat+i-cnt,cnt,0,0,ps_rjitter);
		cnt = 0;
	    }
	    cnt++;
	}
	print_stat_hist(stat+i-cnt,cnt,0,0,ps_rjitter);
	printf("\n");
    }
    if (stat_cumul) {
	int cnt = 0,i,sum = 0;

	printf(
	  "# Cumulative: jitter (%cs), number of occurrences (per %d us)\n",
	  unit_prefix,bin_us);
	for (i = 0; i != n; i++) {
	    if (i && stat[i].rdev_us/bin_us != stat[i-1].rdev_us/bin_us) {
		print_stat_hist(stat+i-cnt,cnt,1,sum,ps_rjitter);
		cnt = 0;
	    }
	    sum++;
	    cnt++;
	}
	print_stat_hist(stat+i-cnt,cnt,1,sum,ps_rjitter);
	printf("\n");
    }
    if (st_opt.stat_overall_hist) {
	int cnt = 0,i;

	qsort(stat,n,sizeof(struct stats),cmp_tdev);
	printf("# Total jitter (%cs), number of occurrences (per %d us)\n",
	  unit_prefix,bin_us);
	for (i = 0; i != n; i++) {
	    if (i && stat[i].tdev_us/bin_us != stat[i-1].tdev_us/bin_us) {
		print_stat_hist(stat+i-cnt,cnt,0,0,ps_tjitter);
		cnt = 0;
	    }
	    cnt++;
	}
	print_stat_hist(stat+i-cnt,cnt,0,0,ps_tjitter);
	printf("\n");
    }
    if (st_opt.stat_lat_hist) {
	int cnt = 0,i;

	qsort(stat,n,sizeof(struct stats),cmp_lat);
	printf("# Latency for %s (%cs), number of occurrences (per %d us)\n",
	  do_write ? "write" : "read",unit_prefix,bin_us);
	for (i = 0; i != n; i++) {
	    if (i && stat[i].lat_us/bin_us != stat[i-1].lat_us/bin_us) {
		print_stat_hist(stat+i-cnt,cnt,0,0,ps_lat);
		cnt = 0;
	    }
	    cnt++;
	}
	print_stat_hist(stat+i-cnt,cnt,0,0,ps_lat);
	printf("\n");
    }
    if (st_opt.stat_pos_hist) {
	int cnt = 0,i;

	qsort(stat,n,sizeof(struct stats),cmp_pos);
	printf("# Advancing (playout point) latency (%cs), "
	  "number of occurrences (per %d us)\n",unit_prefix,bin_us);
	for (i = 0; i != n; i++) {
	    if (i && stat[i].pos_us/bin_us != stat[i-1].pos_us/bin_us) {
		print_stat_hist(stat+i-cnt,cnt,0,0,ps_pos);
		cnt = 0;
	    }
	    cnt++;
	}
	print_stat_hist(stat+i-cnt,cnt,0,0,ps_pos);
	printf("\n");
    }
    if (st_opt.stat_dist_hist) {
	int cnt = 0,i;

	qsort(stat,n,sizeof(struct stats),cmp_dist);
	printf("# Movement distance (effective), number of occurrences\n");
	for (i = 0; i != n; i++) {
	    if (i && stat[i].dist_bytes != stat[i-1].dist_bytes) {
		print_stat_hist(stat+i-cnt,cnt,0,0,ps_dist);
		cnt = 0;
	    }
	    cnt++;
	}
	print_stat_hist(stat+i-cnt,cnt,0,0,ps_dist);
	printf("\n");
    }
    if (seq_errors) {
	fprintf(stderr,"%d sequence error%s\n",seq_errors,
	  seq_errors == 1 ? "" : "s");
	return 1;
    }
    return 0;
}
