/*
 * Script.c - C framework for buffer unit tests
 *
 * Written 2004-2006 by Werner Almesberger
 *
 * (C) Copyright Koninklijke Philips Electronics NV 2004-2006. All rights
 * reserved.
 * For licensing and warranty information, see the file ../../COPYING.GPLv2
 */


#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <assert.h>


#define DPRINTK3(...)
#define DPRINTK4(...)
#define KERN_WARNING ""
#define printk(...)
#define PAGE_SIZE 4096
#define PAGE_MASK (~(PAGE_SIZE-1))
#define GFP_KERNEL 0
#define ENOMEM -1
#define EINVAL -1

/*
 * We can't just do an assert(x), because the problem may not surface at the
 * the first assert. We have to terminate immediately when we hit the problem,
 * because things may get wildly inconsistent.
 */

#define BUG_ON(x) \
  ({ if (expect_failure) { \
       if (x) \
	 _exit(0); \
     } \
     else \
        assert(!(x)); })


static int expect_failure = 0;

struct page {
    int dummy;
};


/*
 * A bit tricky, because we have to maintain page alignment, which malloc
 * doesn't normally do.
 */

static unsigned long __get_free_page(int gfp)
{
    void *page;
    void **aligned;

    page = malloc(PAGE_SIZE*2+sizeof(void *)-1);
    if (!page) {
	perror("malloc");
	exit(1);
    }
    aligned = (void **)
      (((unsigned long) page+PAGE_SIZE+sizeof(void *)-1) & PAGE_MASK);
    aligned[-1] = page;
    return (unsigned long) aligned;
}


static void free_page(unsigned long page)
{
    free(((void **) page)[-1]);
}


#include "_buffer.h"
#include "_buffer.c"


static struct abiss_buffer buffer;
static struct page_slot *ptr;


/* ----- Buffer content ---------------------------------------------------- */


static void __attribute__((used)) resize(int slots)
{
    abiss_resize_buffer(&buffer,slots);
}


static void __attribute__((used)) add(unsigned long page)
{
    
    *abiss_grow_active(&buffer) = (struct page *) page;
}


static void __attribute__((used)) rem(unsigned long page)
{
    assert(abiss_shrink_active(&buffer) == (struct page *) page);
}


static void __attribute__((used)) rem_any(void)
{
    abiss_shrink_active(&buffer);
}


#define resize(n) resize(n);
#define add(p) add(p);
#define rem(p) rem(p);
#define rem_any rem_any();


/* ----- Buffer status ----------------------------------------------------- */


static int __attribute__((used)) fullq(void)
{
    return abiss_buffer_full(&buffer);
}


static void __attribute__((used)) full(void)
{
    assert(fullq());
}


static void __attribute__((used)) nfull(void)
{
    assert(!fullq());
}


static int __attribute__((used)) emptyq(void)
{
    return abiss_buffer_empty(&buffer);
}


static void __attribute__((used)) empty(void)
{
    assert(emptyq());
}


static void __attribute__((used)) nempty(void)
{
    assert(!emptyq());
}


#define fullq fullq()
#define full full();
#define nfull nfull();
#define emptyq emptyq()
#define empty empty();
#define nempty nempty();


/* ----- Pointer movement -------------------------------------------------- */


static void __attribute__((used)) start(void)
{
    ptr = abiss_buffer_start(&buffer);
}


static void __attribute__((used)) next(void)
{
    ptr = next_page_slot(ptr);
}


static void __attribute__((used)) isat(unsigned long page)
{
    assert(ptr->page == (struct page *) page);
}


/*
 * We're actually not "at" the end, but rather beyond it.
 */

static void __attribute__((used)) atend(void)
{
    assert(!ptr);
}


static void __attribute__((used)) natend(void)
{
    assert(ptr);
}


#define start start();
#define next next();
#define isat(p) isat(p);
#define atend atend();
#define natend natend();


/* ----- Main and abort handler -------------------------------------------- */


static void __attribute__((used)) willfail(void)
{
    expect_failure = 1;
}


#define willfail willfail();


int main(void)
{
    abiss_init_buffer(&buffer);

#include SCRIPT_FILE

    return -expect_failure;
}
